(herald "fine/sgx_minimal_with_target") 

(defprotocol sgx-min basic
  (defrole local-quote
    (vars (er mesg) (proc-sec skey) (target name))
    (trace
     (recv (cat er target))
     (send (cat er (hash er (hash "mac" target proc-sec))))))

  (defrole epid-quote
    (vars (er mesg) (k-epid akey) (proc-sec skey) (target name))
    (trace
     (recv (cat er (hash er (hash "mac" target proc-sec))))
     (send (cat er (enc "rq" er k-epid)))))

  (defrole attest-server
    (vars (er mesg) (n text) (k-epid akey) (as name))
    (trace
     (recv (enc n er (enc "rq" er k-epid) (pubk as)))
     (send n)))

  (defrole attest-client
    (vars (m er mesg) (n text) (as name))
    (trace
     (recv (cat er m)) 
     (send (enc n er m (pubk as)))
     (recv n)))

  (defrule as-confirms-epid-ok
    (forall
     ((z-as strd) (k-epid akey))
     (implies (and (p "attest-server" z-as 2)
		   (p "attest-server" "k-epid" z-as k-epid))
	      (and (fact man-made-epid k-epid) ;; AS confirms that EPID was used 
		   (non k-epid)))))   ;; and the signing key
				      ;; uncompromised

  (defrule epid-quote-processor-ok
    (forall
     ((z-iq strd) (proc-sec skey) (k-epid akey))
     (implies (and (p "epid-quote" z-iq 2)
		   (p "epid-quote" "k-epid" z-iq k-epid)
		   (p "epid-quote" "proc-sec" z-iq proc-sec)
		   (fact man-made-epid k-epid))
	      (non proc-sec))))		;; If the EPID quote role ran,
					;; then the associated processor secret is uncompromised

  (defrule local-quote-attests-enclave
    (forall
     ((z-lq strd) (eid ch rest mesg) (k akey) (proc-sec skey))
     (implies (and (p "local-quote" z-lq 2)
		   (p "local-quote" "er" z-lq (cat eid ch k rest))
		   (p "local-quote" "proc-sec" z-lq proc-sec)
		   (non proc-sec))	; If the local quote role runs,
					; with uncompromised processor key,
					; then an enclave exists
	      (fact enclave-w-code-and-key eid ch k proc-sec))))
  )

(defskeleton sgx-min
  (vars (as name) (eid ch rest mesg) (k akey) (n text))
  (defstrand attest-client 3
    (er (cat eid ch k rest)) (n n) (as as))
  (non-orig (privk as))
  (uniq-orig n))

