(herald "cb" (bound 12))

(defprotocol cb-over-sgx basic 
  (defrole local-quote
    (vars (er mesg) (proc-sec skey))
    (trace
     (recv er)
     (send (cat er (hash er proc-sec)))))

  (defrole epid-quote
    (vars (er mesg) (k-epid akey) (proc-sec skey))
    (trace
     (recv (cat er (hash er proc-sec)))
     (send (cat er (enc er k-epid)))))

  (defrole attest-server
    (vars (er mesg) (n text) (k-epid akey) (as name))
    (trace
     (recv (enc n er (enc er k-epid) (pubk as)))
     (send n)))

  (defrole fulcrum
    (vars (m er mesg) (n text) (as f name))
    (trace
     (recv (cat er m)) 
     (send (enc n er m (pubk as)))
     (recv n)
     (send (enc er (privk f)))))

  (defrole certifier
    (vars (ca f name) (k akey))
    (trace
     (send (enc f k (privk ca)))))
 
  (defrole crowbar
    (vars (er mesg) (cbkey akey) (proc-sec skey))
    (trace
     (recv (cat er (hash er proc-sec)))
     (send (enc er (invk cbkey)))))

   (defrule as-confirms-epid-ok
    (forall
     ((z-as strd) (k-epid akey))
     (implies (and (p "attest-server" z-as 2)
		   (p "attest-server" "k-epid" z-as k-epid))
	      (and (fact man-made-epid k-epid) ;; AS confirms that EPID was used 
		   (non k-epid)))))
   
   (defrule as-confirms-non-std
    (forall
     ((z-as strd) (k-epid akey))
     (implies (and (p "attest-server" z-as 2)
		   (p "attest-server" "k-epid" z-as k-epid)
		   (fact is-std-sigkey k-epid))
	      (false))))

   (defrule as-key-std
     (forall ((z-as strd) (as name))
	     (implies (and (p "attest-server" z-as 2)
			   (p "attest-server" "as" z-as as))
		      (fact is-std-sigkey (privk as)))))

  (defrule epid-quote-processor-ok
    (forall
     ((z-iq strd) (proc-sec skey) (k-epid akey))
     (implies (and (p "epid-quote" z-iq 2)
		   (p "epid-quote" "k-epid" z-iq k-epid)
		   (p "epid-quote" "proc-sec" z-iq proc-sec)
		   (fact man-made-epid k-epid))
	      (non proc-sec))))

  ;;; New rule:  key used by epid-quote is epid
  (defrule epid-quote-non-std
    (forall
     ((z-iq strd) (k-epid akey))
     (implies (and (p "epid-quote" z-iq 2)
		   (p "epid-quote" "k-epid" z-iq k-epid)
		   (fact is-std-sigkey k-epid))
	      (false))))

  (defrule local-quote-attests-enclave
    (forall
     ((z-lq strd) (eid ch rest mesg) (k akey) (proc-sec skey))
     (implies (and (p "local-quote" z-lq 2)
		   (p "local-quote" "er" z-lq (cat eid ch k rest))
		   (p "local-quote" "proc-sec" z-lq proc-sec)
		   (non proc-sec))	; If the local quote role runs,
					; with uncompromised processor key,
					; then an enclave exists
	      (fact enclave-w-code-and-key eid ch k proc-sec))))  
  

  (defrule cert-rule-fulcrum
    ;; cert vouches for uncompromised key 
    (forall
     ((z-ca strd) (f name) (k akey))
     (implies (and (p "certifier" z-ca 1)
		   (p "certifier" "k" z-ca k)
		   (p "certifier" "f" z-ca f)
		   (= k (pubk f)))	; only assuming pubk to agree
					; with cert
	      (non (invk k)))))

    ;;; New rule:  CA key non epid
  (defrule cert-non-epid 
    (forall
     ((z-ca strd) (ca name))
     (implies (and (p "certifier" z-ca 1)
		   (p "certifier" "ca" z-ca ca))
	      (fact is-std-sigkey (privk ca)))))

  ;;; New rule:  fulcrum key non epid
  (defrule fulcrum-non-epid
    (forall ((z-fm strd) (f name))
	    (implies (and (p "fulcrum" z-fm 2)
			  (p "fulcrum" "f" z-fm f))
		     (fact is-std-sigkey (privk f)))))

  ;;; New rule:  "trust-anchor" property guarantees non
  (defrule trust-anchor-inverse-is-non
    (forall ((k akey))
	    (implies (fact trust-anchor k)
		     (non (invk k)))))

  ;;; New rule:  "trust-anchor" not used by fulcrum
  (defrule fulcrum-doesnt-use-trust-anchor
    (forall ((z strd)(f name))
	    (implies (and (fact trust-anchor (pubk f))
			  (p "fulcrum" z 2)
			  (p "fulcrum" "f" z f))
		     (false))))

  ;;; New rule:  "trust-anchor" not used by crowbar
  (defrule crowbar-doesnt-use-trust-anchor
    (forall ((z strd)(cbkey akey))
	    (implies (and (fact trust-anchor cbkey)
			  (p "crowbar" z 2)
			  (p "crowbar" "cbkey" z cbkey))
		     (false))))
  

  (defrule crowbar-doesnt-use-fulcrum-key
    (forall ((z z-ca strd)(f name))
	    (implies (and (p "certifier" z-ca 1)
			  (p "certifier" "f" z-ca f)
			  (p "crowbar" z 2)
			  (p "crowbar" "cbkey" z (pubk f)))
		     (false))))

  (defrule cert-rule-attest-server
    ;; fulcrum will select a valid attest server
    ;; and a fresh challenge nonce.  
    (forall ((z-fm strd)(as name) (n text))
	    (implies (and (p "fulcrum" z-fm 2)
			  (p "fulcrum" "as" z-fm as)
			  (p "fulcrum" "n" z-fm n))
		     (and (non (privk as))
			  (uniq n)))))

  (defrule local-att-crowbar
    ;; cbkey resident in a valid enclave
    ;; controlled by crowbar code
    ;; successfully protects its private key
    (forall ((eid chc mesg) (cbkey akey) (proc-sec skey))
  	    (implies (and (fact enclave-w-code-and-key eid chc cbkey proc-sec)
  			  (fact known-crowbar-code chc))
  		     (non (invk cbkey)))))

  (defrule crowbar-std-sig
    ;; a valid enclave controlled by crowbar code
    ;; will not migrate its private key to a new processor 
    (forall ((z strd) (cbkey akey))
	    (implies (and (p "crowbar" z 2)
			  (p "crowbar" "cbkey" z cbkey))
		     (fact is-std-sigkey (invk cbkey)))))
  
  (defrule crowbar-immobile
    ;; a valid enclave controlled by crowbar code
    ;; will not migrate its private key to a new processor 
    (forall ((z strd) (eid chc mesg) (proc-sec proc-sec-0 skey) (cbkey akey))
	    (implies (and (fact enclave-w-code-and-key eid chc cbkey proc-sec)
			  (fact known-crowbar-code chc)
			  (p "crowbar" z 2)
			  (p "crowbar" "cbkey" z cbkey)
			  (p "crowbar" "proc-sec" z proc-sec-0))
		     (= proc-sec proc-sec-0))))

  (defrule crowbar-code-never-ca
    (forall ((z-ca strd) (eid chc mesg) (proc-sec skey) (ca name))
	    (implies (and (fact enclave-w-code-and-key eid chc (pubk ca) proc-sec)
			  (fact known-crowbar-code chc)
			  (p "certifier" z-ca 1)
			  (p "certifier" "ca" z-ca ca))
		     (false))))

  
  (defrule crowbar-code-never-fulcrum
    (forall ((z-f strd) (eid chc mesg) (proc-sec skey) (f name))
	    (implies (and (fact enclave-w-code-and-key eid chc (pubk f) proc-sec)
			  (fact known-crowbar-code chc)
			  (p "fulcrum" z-f 4)
			  (p "fulcrum" "f" z-f f))
		     (false))))

  (defrule neq-def
    ;; logical rule for negated equality 
    (forall ((m mesg))
	    (implies (fact neq m m)
		     (false))))

  (defrule sig-keys-inverse-std 
    (forall ((k akey))
	    (implies (fact is-std-sigkey k)
		     (fact is-std-sigkey (invk k))))))

(defskeleton cb-over-sgx
  (vars (f ca name))
  (deflistener (enc f (pubk f) (privk ca)))
  (facts (trust-anchor (pubk ca))
	 (is-std-sigkey (privk ca))))

(defskeleton cb-over-sgx
  (vars (f ca name) (k akey) (eid ch rest mesg))
  (deflistener (cat (enc f (pubk f) (privk ca))
		    (enc (cat eid ch k rest) (privk f))))
  (facts (trust-anchor (pubk ca))
	 (is-std-sigkey (privk ca))
	 (is-std-sigkey (privk f))))

(defskeleton cb-over-sgx
  (vars (f ca name) (cbkey akey) (er eid chc rest mesg))
  (deflistener (cat (enc f (pubk f) (privk ca))
		    (enc (cat eid chc cbkey rest) (privk f))
		    (enc er (invk cbkey))))
  (facts  (trust-anchor (pubk ca))
	  (is-std-sigkey (privk ca))
	  (is-std-sigkey (privk f))
	  (is-std-sigkey cbkey)
	  (known-crowbar-code chc)
	  (neq (privk f) (invk cbkey))))

(defskeleton cb-over-sgx
  (vars (f ca name) (cbkey akey) (eidc chc er rest mesg))
  (deflistener (cat (enc f (pubk f) (privk ca))
		    (enc (cat eidc chc cbkey rest) (privk f))
		    (enc er (invk cbkey))))
  (facts (trust-anchor (pubk ca))
	 (is-std-sigkey (privk ca))
	 (is-std-sigkey (privk f))
	 (is-std-sigkey cbkey)
	 (known-crowbar-code chc)
	 (neq er (cat eidc chc cbkey rest))))


(defskeleton cb-over-sgx
  (vars (f ca name) (cbkey ka akey) (eidc eida chc cha rest resta mesg))
  (deflistener (cat (enc f (pubk f) (privk ca))
		    (enc (cat eidc chc cbkey rest) (privk f))
		    (enc (cat eida cha ka resta) (invk cbkey))))
  (facts (trust-anchor (pubk ca))
	 (is-std-sigkey (privk ca))
	 (is-std-sigkey (privk f))
	 (is-std-sigkey cbkey)
	 (known-crowbar-code chc)
	 (neq chc cha)))

